
local VERSION = 1.21

--[[### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####
]]

-- @author Michael Martin with much appreciated code by Thomas Loockx, Roeland,
--                          gmillas, Jason Grimes, Mark Bassett & Matej Moderc.
-- @shortcut alt + a
--
-------------------------------------------------------------------------------
--
--                       Ol' Ready Cam (ORC)
--                 Orc File / Key frame creation script
--
-- Note: When changing shortcut keys for these Lua scripts, open
-- Octane and do the following:
--    1. Click on the 'Script' Menu.
--    2. Select 'Reload Scripts'
-- Shortcuts can be changed on line 22 above.
-- Use either alt or ctrl plus '+' and a letter.
-- The two steps above need to be done whenever shortcuts are changed or
-- new scripts added. Make sure the keys don't conflict with other scripts.
--
-- Summary:
-- This script is used for creating Orc files within Octane. It can be
-- executed with the shortcut 'Alt A'. When executed, the following occurs:
--    I. Based on the name of the current project, the script generates
--       the name of an Orc File and then...
--   II. A dialog box appears asking whether you want to
--       1. 'Add Frame' 
--           If you click on 'Add frame', the current camera position is
--           added to the file and the script terminates.
--       2. 'Exit'
--           The script terminates without having done anything.
--       3. 'Clear All Frames'
--           All data that may be in the Orc file is deleted and
--           the script terminates.
--
-- That's it. Move the camera, press 'Alt A', 'Add Frame' and repeat
-- or rinse as often as desired.
--
-- N.B.
-- If you have multiple camera nodes and/or render targets,
-- you should first click on the render target whose camera will be used.
-- Otherwise, the script may default to a camera which isn't the one whose
-- viewport is currently active. The camera position that is saved is of the
-- currently selected target which must have a camera node and kernel node attached.
--
-- When you are done, run the Orc Animation script which can be executed with the
-- shortcut 'Alt R'.
--    The default Orc file of the current Octane project is automatically loaded.
--    i.e. An Octane .ocs file must be present for the script to determine
--    the current project directory and name.
--

------------------------------------------------------------------------------
-- Load ancillary files
------------------------------------------------------------------------------
require "Orc_bin/gui"
require "Orc_bin/dialog"
require "Orc_bin/render"
require "Orc_bin/scene"

--------------------------------------------------------------------------------
-- Save the current Octane camera position to the default Orc file.
--------------------------------------------------------------------------------
local function saveCameraPosition()

    local idx, currentTarget = scene:getRenderTarget()
    local camNode = currentTarget:getInputNode(octane.P_CAMERA)

    local function pinValue(c,op)
        local v = c:getPinValue(op)
        return string.format("%s,%s,%s", v[1], v[2], v[3])
    end

    local renderTarget = octane.project.getSelection()[1]
    if renderTarget ~= nil then
        if renderTarget.type ~= octane.nodeType.NT_RENDERTARGET then
            renderTarget = nil
        end
    end

    local x = camNode:getProperties()
    if x.type ~= octane.NT_CAM_PANORAMIC and 
        x.type ~= octane.NT_CAM_THINLENS then
        local msg = "No thinlens or panoramic camera\nfound attached to render target.\n"
        dialog.warning("ORC Key Framing terminated", msg)
    else
        local t = pinValue(camNode, octane.P_TARGET)
        local p = pinValue(camNode, octane.P_POSITION)
        local u = pinValue(camNode, octane.P_UP)
        local s = string.format("%s,%s,%s\n", t,p,u)

        local msg1 = ""
        local msg2 = ""
        if renderTarget == nil and #Render.targets > 1 then
            msg1 = "ORC Key Framing\n\nThere are multiple render targets\n"
            msg1 = msg1 .. "and none are selected...\n"
            msg1 = msg1 .. "defaulting to render target:\n"..currentTarget.name
            msg2 = "Select an option for camera positioning in\n"..cameraFile.editor.text
        else
            msg1 = "ORC Key Framing"
            msg2 = "Select an option for camera positioning. in\n"..cameraFile.editor.text
        end

        local answer = dialog.action(msg1, msg2)
        if answer == 1 then
            local file = io.open(cameraFile.editor.text, "a")
            if file ~= nil then
                local vtype = file:write(s)
                file:close()
            end
        elseif answer == 3 then
            local file = io.open(cameraFile.editor.text, "w")
            if file ~= nil then
                file:close()
            end
        end
    end
end

--------------------------------------------------------------------------------
-- Initialize the program enough for handling the writing of camera data
-- into the default Orc file.
--------------------------------------------------------------------------------
local function initializeFrameAdding()

    -- Make sure we've got at least one render target with a kernel node.
    -- What's available is loaded in renderTargets.combo
    if not scene:initializeRenderTargets() then
        return false
    end

    -- Get the render target and camera in global variables
    if scene:getCopy() == false then
       return false
    end

    -- Get the Octane project path and name and swap the ocs extension with orc
    Render.projectPath = Render.projectPath:gsub("\\", "/")

    if octane.project.getCurrentProject() == "" then
        local msg = "No Octane project exists for the current layout.\nFirst save your work as an Octane project\n"
        msg = msg ..  "before trying to add ORC frames."
        dialog.error("ORC Key Framing Error", msg)
        return false
    end

    local orcFile = scene:getDefaultOrcFile()
    
    cameraFile.editor:updateProperties { text=orcFile }

    local file = io.open(cameraFile.editor.text, "r")
    if file ~= nil then
        file:close()
    end

    return true
end

--------------------------------------------------------------------------------
-- Start here for saving camera position
--------------------------------------------------------------------------------
if( initializeFrameAdding()) then
    saveCameraPosition()
end
