
-- :::::::::::::::::::::::::::::
-- ::                         ::
-- ::      KeyFrame Tab       ::
-- ::                         ::
-- :::::::::::::::::::::::::::::...............................................

keyframe = widget:create()
keyframe.path = ""
keyframe.activeButton = nil -- points to render button currently active whether
                            -- for individual keys or previewing all data
keyframe.orcCache = {}      -- All the camera records generated from the cameraFile
                            -- including all frames per key after bezier smoothing.
keyframe.rebuildOrcCache = true

--  These values can be preserved in the Orc_Defaults file
keyframe.maxSliderValue     = settings.keyframeMax.numeric.value
keyframe.maxSamples         = settings.keyframeSamples.numeric.value
keyframe.defaultSliderValue = settings.keyframeValue.numeric.value
keyframe.screenLimitDefault = settings.keyframeLimit.numeric.value+1
keyframe.screenLimit        = keyframe.screenLimitDefault 

keyframe:new('button','saveAs')      {"Save As...", width=110, x=20,
                                      tooltip="Save interpolated frames as ORC file."}
keyframe:new('numeric','smoothing')  {{"Bezier Smoothing", width=110, x=80}, width=80,
                                       value=.5, min=0.0, max=1.0, step=0.1 }

keyframe:new('slider','showFrame')   {{"Show Frame", width=90, x=6}, width=410}
keyframe:new('button','prev')        {"Previous"}
keyframe:new('button','next')        {"Next"}
keyframe:new('slider','samples')     {{"Samples/px", width=80}, width=410,max=keyframe.maxSamples,log=true}
keyframe:new('slider','shutterTime') {{"Shutter Time",width=80}, log=true,
                                        width=410, min=0, max=2, step=0.001,
                                        value=settings.shutterTime.numeric.value}

--  ..............................................................................

keyframe:new('label', 'header')  {"Key Frame      Frames in Key", width=170, x=6}
keyframe:new('check', 'lock')    {text="Lock sliders"}
keyframe:new('button','preview') {"Preview", width=110, height=24, x=6,
                                  tooltip="Run preview of interpolated frames."}
--  ..............................................................................

--  ...........................................................................
--             Define sliders and buttons for each key frame key
--  ...........................................................................
keyframe.screenItem  = {}
for i=1,keyframe.screenLimit do
    keyframe.screenItem[i] = keyframe:new('slider') {{"Key "..i, width=40}, width=410,
        height=20, value=keyframe.defaultSliderValue, max=keyframe.maxSliderValue }
    keyframe.screenItem[i]:append('button') {"Key "..i, height=20}
end

keyframe.item = {}    -- Each item corresponds to a cameraFile.data record and contains
                      -- the number of interpolated records to be produced for a key.
keyframe.index  = 1   -- Index into cameraFile.data that is currently the 1st item
                      -- in keyframe.screenItem. For example, 500 means that the 500th
                      -- record is cameraFile.data[500] in keyframe.screenItem[1].
keyframe.totalFrames = 0  -- Total number of interpolated frames

--  ...........................................................................


--------------------------------------------------------------------------------------
-- Screen groups and layout for the Keyframe tab
--------------------------------------------------------------------------------------
do
    local children =
    { 
        {  keyframe.showFrame.label, keyframe.showFrame.slider }
    }
    local g1 = keyframe:new('group') { children, padding={0,2}  }

    ---------------------------------------------------------------------
    children =
    { 
        {keyframe.header.label, keyframe.lock.check, keyframe.preview.button }
    }
    local g2 = keyframe:new('group') { children, padding={4,4} }

    ---------------------------------------------------------------------
    children = {}
    for i,v in pairs(keyframe.screenItem) do
        local item = { v.button, v.slider }
        table.insert(children, item)
        if i == keyframe.screenLimit-1 then
            break
        end
    end
    local g3 = keyframe:new('group') { children, padding={5,1}, inset={2} }

    ---------------------------------------------------------------------
    children =
    {
        { keyframe.prev.button, keyframe.next.button, keyframe.smoothing.label, keyframe.smoothing.numeric },
    }
    local g4 = keyframe:new('group') { children, padding={5,4}, x=2 }

    ---------------------------------------------------------------------
    children =
    {
        { keyframe.samples.label, keyframe.samples.slider },
        { keyframe.shutterTime.label, keyframe.shutterTime.slider }
    }
    local g5 = keyframe:new('group') { children, padding={6,2} }

    ---------------------------------------------------------------------
    children =
    {
        { g1.group },
        { g2.group },
        { g3.group },
        { g4.group },
        { g5.group }
    }
    keyframe:append('group') { children, border=true, padding={10,2}, inset={10} }
end ----------------------------------------------------------------------------


--------------------------------------------------------------------------------
-- Keyframe functions //////////////////////////////////////////////////////////
--------------------------------------------------------------------------------
function keyframe:loadScreenItems()

    local j = 1
    for i=self.index, self.index+(self.screenLimit-1) do
        local si = self.screenItem[j]
        if i >= #self.item then
            if si ~= nil then
                si.button.text = ""
                si.slider.value = self.defaultSliderValue
                si.slider.enable = false
                si.button.enable = false
            end
        else
            si.button.text = "k: " .. i
            si.slider.value = self.item[i]
            si.slider.enable = true
            si.button.enable = true
        end
        j = j+1
    end
end

-------------------------------------------------------------------
function keyframe:enable()
 
    local enabler = cameraFile.editor.text ~= ""

    self.showFrame.slider:updateProperties   { enable=enabler }
    self.lock.check:updateProperties         { enable=enabler }
    self.smoothing.numeric:updateProperties  { enable=enabler }
    self.samples.slider:updateProperties     { enable=enabler }
    self.shutterTime.slider:updateProperties { enable=enabler }
    self:loadScreenItems()
    self.saveAs.button:updateProperties      { enable=enabler }
    self.preview.button:updateProperties     { enable=enabler }
    self.prev.button:updateProperties        { enable=enabler }
    self.next.button:updateProperties        { enable=enabler }
end


-------------------------------------------------------------------
function keyframe:disable()

    self.showFrame.slider:updateProperties   { enable=false }
    self.lock.check:updateProperties         { enable=false }
    self.smoothing.numeric:updateProperties  { enable=false }
    self.samples.slider:updateProperties     { enable=false }
    self.shutterTime.slider:updateProperties { enable=false }
    self.preview.button:updateProperties     { enable=false }
    self.saveAs.button:updateProperties      { enable=false }
    self.prev.button:updateProperties        { enable=false }
    self.next.button:updateProperties        { enable=false }
    for i=1, self.screenLimit-1 do
        self.screenItem[i].label:updateProperties  { enable=false }
        self.screenItem[i].button:updateProperties { enable=false }
        self.screenItem[i].slider:updateProperties { enable=false }
    end
end


--------------------------------------------------------------------------------
function keyframe:setDefaultSliderValues()

    for i=1, cameraFile.totalFrames do
        keyframe.item[i] = keyframe.defaultSliderValue
    end

    local ix = keyframe.index
    for i=1, keyframe.screenLimit-1 do 
        if keyframe.item[ix] then
            keyframe.screenItem[i].slider.value = keyframe.item[ix]
        end
        ix = ix+1
    end
end

--------------------------------------------------------------------------------
function keyframe:updateSliderValues(diff, sIdx)

     -- Get item index of screen item
    local itemIdx = keyframe.index + (sIdx - 1)

    keyframe.totalFrames = 0
    for i=1, cameraFile.totalFrames-1 do -- Update all item values except the current
        if i ~= itemIdx then             -- one which was the actual slider moved.
            keyframe.item[i] = keyframe.item[i] + diff
            if keyframe.item[i] < 1 then
                keyframe.item[i] = 1
            elseif keyframe.item[i] > keyframe.maxSliderValue then
                keyframe.item[i] = keyframe.maxSliderValue
            end
        end
        keyframe.totalFrames = keyframe.totalFrames + keyframe.item[i]
    end

    -- Get the screen items to match the stored values
    local ix = keyframe.index
    for i=1, keyframe.screenLimit-1 do 
        if ix < cameraFile.totalFrames then
            keyframe.screenItem[i].slider.value = keyframe.item[ix]
        end
        ix = ix+1
    end
 end
 
--------------------------------------------------------------------------------
-- The number of slots in keyframe.item[] matches the number of cameraFile.data
-- i.e. Orc records. Each Orc record is a keyframe. The value held in each
-- keyframe.item[] slot is the number of interpolated frames that are to be 
-- produced for that keyframe. This value is originally set to the
-- keyframe.defaultSliderValue and is only changed when its corresponding
-- slider is moved on the screen.
--------------------------------------------------------------------------------
function keyframe:initializeData()

    keyframe.item = {}
    keyframe.totalFrames = 0  -- Total number of interpolated frames
    keyframe.index = 1        -- Index into keyframe.item[] that is to be the 1st
                              -- item appearing in keyframe.screenItem[1]
    for i=1, cameraFile.totalFrames do
        keyframe.item[i] = keyframe.defaultSliderValue
        if i < cameraFile.totalFrames then
            keyframe.totalFrames = keyframe.totalFrames + keyframe.defaultSliderValue
        end
    end

    -- maxValue should always show the total number of interpolated frames
    keyframe.showFrame.slider.maxValue = keyframe.totalFrames
    animate.startFrame.slider.maxValue = keyframe.totalFrames
    animate.endFrame.slider.maxValue   = keyframe.totalFrames
    if animate.startFrame.slider.value > animate.startFrame.slider.maxValue then
        animate.startFrame.slider.value = 1
    end
    if animate.endFrame.slider.value >  animate.endFrame.slider.maxValue then
        animate.endFrame.slider.value =  animate.endFrame.slider.maxValue
    end

    -- More sliders can appear on the screen by changing 
    -- the keyframe.screenLimitDefault
    keyframe.screenLimit = keyframe.screenLimitDefault

    local index = keyframe.index

    for i=1, keyframe.screenLimit-1 do
        if index < cameraFile.totalFrames then
            keyframe.screenItem[i].button.text = "k: "..index
            keyframe.screenItem[i].button:updateProperties {enable=true}
            keyframe.screenItem[i].slider:updateProperties {enable=true, value=keyframe.defaultSliderValue}
        else
            keyframe.screenItem[i].button.text = ""
            keyframe.screenItem[i].button:updateProperties {enable=false}
            keyframe.screenItem[i].slider:updateProperties {enable=false, value=keyframe.defaultSliderValue}
        end
        index = index + 1
    end

    if cameraFile.totalFrames < keyframe.screenLimit then
        keyframe.screenLimit = cameraFile.totalFrames
    end
    
    keyframe.prev.button:updateProperties {enable=true}
    keyframe.next.button:updateProperties {enable=true}

    keyframe.showFrame.slider.value = 1
    keyframe.rebuildOrcCache = true
end

--------------------------------------------------------------------------------
function keyframe:unpackFrameData(frame)

    local c = keyframe.orcCache[frame]
    local t = {c[1],c[2],c[3]}
    local p = {c[4],c[5],c[6]}
    local u = nil
    if c[7] ~= nil then
        u = {c[7],c[8],c[9]}
    else
        u = {0,1,0}
    end

    return t, p, u
end

--------------------------------------------------------------------------------
-- From: http://www.fractalnet.org/CAD/curve/BezierCurve.pdf
--------------------------------------------------------------------------------
local function Bezier3(p1, p2, p3, mu)
    local mu2 = mu * mu
    local mum1 = 1 - mu
    local mum2 = mum1 * mum1
    local p = {}
    p[1] = p1[1] * mum2 + 2 * p2[1] * mum1 * mu + p3[1] * mu2
    p[2] = p1[2] * mum2 + 2 * p2[2] * mum1 * mu + p3[2] * mu2
    p[3] = p1[3] * mum2 + 2 * p2[3] * mum1 * mu + p3[3] * mu2
    return p
end

function getVectorDistance(p1, p2)
    local xd = p2[1] - p1[1]
    local yd = p2[2] - p1[2]
    local zd = p2[3] - p1[3]
    local v = (xd*xd + yd*yd + zd*zd)
    local d = math.sqrt(v)

    return d
end  

--------------------------------------------------------------------------------
-- This function creates a cache of bezier smoothed cameraFile records.
-- If the number of frames of any key have been modified, this cache will be
-- regenerated when any action requiring a rendered frame is initiated.
--------------------------------------------------------------------------------
function keyframe:generateOrcCache()

    keyframe.rebuildOrcCache = false

    -- We need at least 2 frames for interpolating
    if cameraFile.totalFrames < 2 then
        return false
    end

    keyframe.orcCache = {}
    local cpCache = {}
    local row = 0
    local maxKeys = cameraFile.totalFrames

    for key = 1, maxKeys-1 do
        local t1, p1, u1 = cameraFile:unpackFrameData(key)
        local t2, p2, u2 = cameraFile:unpackFrameData(key+1)

        local frameTotal = keyframe.item[key]
        table.insert(keyframe.orcCache, cameraFile.data[key])
        row = row + 1
        table.insert(cpCache, row)

        for f=1, frameTotal-1 do
            local mu = 1.0000 / frameTotal * f
            local t = octane.vec.lerp(t1, t2, mu)
            local p = octane.vec.lerp(p1, p2, mu)
            local u = octane.vec.lerp(u1, u2, mu)
            row = row + 1
            local r = {}
            for i=1,3 do
                r[i]   = t[i]
                r[i+3] = p[i]
                r[i+6] = u[i]
            end
            table.insert(keyframe.orcCache, r)
        end
    end            

    if keyframe.showFrame.slider.value > #keyframe.orcCache then
        keyframe.showFrame.slider.value = #keyframe.orcCache
    end

    -- If no smoothing, then just do staight interpolation
    if keyframe.smoothing.numeric.value == 0 then
        return
    end
    
    for key = 1, maxKeys-2 do

        local seg1Total = keyframe.item[key]
        local bez1Total = math.floor(seg1Total * keyframe.smoothing.numeric.value)
        if bez1Total < 1 then bez1Total = 1 end
        local bezStart = cpCache[key+1] - bez1Total

        local seg2Total = keyframe.item[key+1]
        local bez2Total = math.floor(seg2Total * keyframe.smoothing.numeric.value)
        if bez2Total < 1 then bez2Total = 1 end
        local bezStart = cpCache[key+1] - bez1Total
        local bezFinish = cpCache[key+1] + bez2Total

        local t1, p1, u1 = keyframe:unpackFrameData(bezStart)
        local t2, p2, u2 = cameraFile:unpackFrameData(key+1)
        if bezFinish > #keyframe.orcCache then
            bezFinish = #keyframe.orcCache
        end
        local t3, p3, u3 = keyframe:unpackFrameData(bezFinish)
        local bezTotal = bezFinish - bezStart
        local row = bezStart+1
        for f=1, bezTotal-1 do
            local mu = 1.0000 / bezTotal * f
            local tn = Bezier3(t1, t2, t3, mu)
            local pn = Bezier3(p1, p2, p3, mu)
            local un = Bezier3(u1, u2, u3, mu)
            local r = {}
            for i=1,3 do
                r[i]   = tn[i]
                r[i+3] = pn[i]
                r[i+6] = un[i]
            end
            keyframe.orcCache[row] = r
            row = row + 1
        end
    end
end

--------------------------------------------------------------------------------
-- This function will create an Orc file that makes interpolated frames real.
--------------------------------------------------------------------------------
local function saveInterpolatedFile()

    local function vectorToString(v)
        return string.format("%s,%s,%s", v[1], v[2], v[3])
    end

    -- We need at least 2 frames for interpolating
    if cameraFile.totalFrames < 2 then
        return false
    end

    local name = keyframe.path
    name = name:gsub("orc", "csv")
    local file, status = io.open(name, "w")
    if status then
        local msg = "Unable to write to:\n"..name
        msg = msg .. "Status: " .. status
        dialog.critical(msg, false)
        return false
    end

    keyframe:generateOrcCache()
    local counter  = 1
    local maxKeys = cameraFile.totalFrames-1

    local t1, p1, u1 = keyframe:unpackFrameData(1)

    for key=1, #keyframe.orcCache do
        local t1, p1, u1 = keyframe:unpackFrameData(key)
        xs = string.format("0.100000 0.000000 0.000000 %s", p1[1])
        ys = string.format("0.000000 0.100000 0.000000 %s", p1[2])
        zs = string.format("0.000000 0.000000 0.100000 %s", p1[3])
        s  = string.format("%s %s %s\n", xs, ys, zs)
        file:write(s)
        counter = counter + 1
    end
    -- Write out just keyframe part 
    --[[
    for key = 1, cameraFile.totalFrames do
        t,p,u = cameraFile:unpackFrameData(key)
        xs = string.format("0.150000 0.000000 0.000000 %s", p[1])
        ys = string.format("0.000000 0.150000 0.000000 %s", p[2])
        zs = string.format("0.000000 0.000000 0.150000 %s", p[3])
        s  = string.format("%s %s %s\n", xs, ys, zs)
        file:write(s)
    end
    ]]

    file:flush()
    file:close()
    local msg = string.format("Wrote %d records using %d key frames to\n\n%s",
                        counter, cameraFile.totalFrames, name)
    dialog.info("Interpolation completed", msg)
end

--------------------------------------------------------------------------------
-- Keyframe gui callbacks
--------------------------------------------------------------------------------
function keyframe:setupCallbacks()

    -------------------------------------------------
    keyframe.samples.slider.callback = function()
        Render:cancel()
    end

    -------------------------------------------------
    keyframe.shutterTime.slider.callback = function()
        project.changed = true
    end

    -------------------------------------------------
    keyframe.showFrame.slider.callback = function()
        keyframe:renderFrame()
    end

    -------------------------------------------------
    keyframe.smoothing.numeric.callback = function()
        keyframe.rebuildOrcCache = true
    end

    -------------------------------------------------
    local function cancelKeyFramePreview()
        Render:cancel()
        Render.is_keyframe = false
    end

    -------------------------------------------------
    local function groupEnabler(onOff)
        solar.group:updateProperties    { enable=onOff }
        animate.group:updateProperties  { enable=onOff }
        names.group:updateProperties    { enable=onOff }
        settings.group:updateProperties { enable=onOff }
        project.group:updateProperties  { enable=onOff }
    end

    -------------------------------------------------
    local function previewAllKeyFrames()
        Render:cancel()
        if keyframe.rebuildOrcCache then
            keyframe:generateOrcCache()
        end

        keyframe.preview.button.callback = cancelKeyFramePreview
        keyframe.preview.saveTip = keyframe.preview.button.tooltip
        if widget.hideToolTips == false then
            keyframe.preview.button.tooltip = " "
        end
        keyframe.preview.button.text    = "Stop Preview"
        if keyframe.showFrame.slider.value >= keyframe.showFrame.slider.maxValue then
            keyframe.showFrame.slider.value = 1
        end
        gui:disable()
        keyframe.activeButton = keyframe.preview.button
        keyframe.preview.button:updateProperties { enable=true }
        groupEnabler(false)

        Render.is_running = true
        Render.is_keyframe = true

        Render:start(animate.path.editor.text, keyframe.showFrame.slider.value,
                    keyframe.showFrame.slider.maxValue, 1, false)

        Render.is_running = false
        Render.is_keyframe = false
        keyframe.activeButton = nil

        keyframe.preview.button.text = "Preview"
        keyframe.preview.button.tooltip  = keyframe.preview.saveTip
        keyframe.preview.button.callback = previewAllKeyFrames

        gui:enable()
        groupEnabler(true)
    end

    --/////////////////////////////////////////////////////////////////////////

    local function previewThisKeyFrame(key)
        Render:cancel()
        if keyframe.rebuildOrcCache then
            keyframe:generateOrcCache()
        end

        keyframe.screenItem[key].button.callback = cancelKeyFramePreview
        keyframe.screenItem[key].button.text = "Stop Preview"

        gui:disable()
        keyframe.screenItem[key].button:updateProperties {enable=true}
        keyframe.activeButton = keyframe.screenItem[key].button
        groupEnabler(false)

        local counter = 0
        keyframe.showFrame.slider.value = 1

        local keyLimit = keyframe.index + (key-1)
        for j=1, keyLimit do 
            counter = keyframe.item[j]
            if j < keyLimit then
                keyframe.showFrame.slider.value =
                    keyframe.showFrame.slider.value + counter
            end
        end

        local limit = keyframe.showFrame.slider.value + (counter-1)

        Render.is_running = true
        Render.is_keyframe = true

        Render:start(animate.path.editor.text,
                     keyframe.showFrame.slider.value, limit, 1, false)

        Render.is_running = false
        Render.is_keyframe = false
        keyframe.activeButton = nil

        keyframe.screenItem[key].button.text = "k: "..keyframe.index+(key-1)
        keyframe.screenItem[key].button.callback = function() previewThisKeyFrame(key) end
        gui:enable()
        groupEnabler(true)
    end

    ---------------------------------------------------------------

    keyframe.preview.button.callback = previewAllKeyFrames

    ---------------------------------------------------------------

    keyframe.next.button.callback = function()
        if #keyframe.item < keyframe.screenLimit then
            return false
        end
        local n = keyframe.index + (keyframe.screenLimit-1)
        if n < #keyframe.item then
            keyframe.index = n
        else
            keyframe.index = 1
        end
        keyframe:loadScreenItems()
    end

    ----------------------------------------------
    keyframe.prev.button.callback = function()
        if keyframe.index == 1 then
            keyframe.index = #keyframe.item - (keyframe.screenLimit-1)
        else
            keyframe.index = keyframe.index - (keyframe.screenLimit-1)
        end
        if keyframe.index < 1 then
            keyframe.index = 1
        end
        keyframe:loadScreenItems()
    end

    ----------------------------------------------
    keyframe.saveAs.button.callback = function()
        Render:cancel()
        -- choose an output file for interpolated frames
        local ret = octane.gui.showDialog
        {
            type      = octane.gui.dialogType.FILE_DIALOG,
            title     = "Choose the ORC output name",
            path      = Render.projectPath,
            wildcards = "*.orc",
            save      = true,
        }

        -- if a file is chosen
        if ret.result ~= "" then
            local name = ret.result:gsub("\\", "/")
            local status, fullname, filename, ext = names.validate:pathForSave(name, "orc")
            if status == false then
                return false
            end

            if ext == "" then  -- if ext is blank, no check was made for existence
                local file = io.open(fullname)
                if file ~= nil then
                    file:close()
                    local msg = filename.."\nalready exists."
                    local answer = dialog.question(2, msg, "Do you want to replace it?")
                    if answer ~= 1 then
                        return false
                    end
                end
            end

            keyframe.path = fullname
            gui:enable()
            saveInterpolatedFile()
            return true
        else
            keyframe.path = ""
        end

        return false
    end 

    ---------------------------------------------------------------
    -- Remember the slider value for each keyframe item
    ---------------------------------------------------------------
    local function updateKeyFrameItemValue(key)
        local diff = keyframe.item[keyframe.index + (key-1)]
        local ndiff = 0
        if diff < keyframe.screenItem[key].slider.value then
            ndiff = keyframe.screenItem[key].slider.value - diff
            keyframe.totalFrames = keyframe.totalFrames + ndiff
        else
            ndiff = (diff - keyframe.screenItem[key].slider.value) * -1
            keyframe.totalFrames = keyframe.totalFrames + ndiff
        end
        keyframe.item[keyframe.index + (key-1)] = keyframe.screenItem[key].slider.value
        local max = keyframe.totalFrames > 2 and keyframe.totalFrames or 3

        keyframe.showFrame.slider:updateProperties { maxValue=max }
        animate.startFrame.slider:updateProperties { maxValue=max }
        animate.endFrame.slider:updateProperties   { maxValue=max }

        return ndiff
    end


    ---------------------------------------------------------------
    -- Initialize callback for each key frame screen item
    ---------------------------------------------------------------
    for i=1, keyframe.screenLimit-1 do 
        keyframe.screenItem[i].slider.callback =  function()
            project.changed = true
            keyframe.rebuildOrcCache = true
            local diff = updateKeyFrameItemValue(i)

            if keyframe.lock.check.checked == true then
                keyframe:updateSliderValues(diff, i)
                local max = keyframe.totalFrames > 2 and keyframe.totalFrames or 3
                keyframe.showFrame.slider:updateProperties { maxValue=max }
                animate.startFrame.slider:updateProperties { maxValue=max }
                animate.endFrame.slider:updateProperties   { maxValue=max }
            end
        end
        keyframe.screenItem[i].button.callback = function()
            previewThisKeyFrame(i)
        end
    end
end

--------------------------------------------------------------------------------
--                  Render an interpolated keyframe
--------------------------------------------------------------------------------
local function renderKeyFrame(t,p,u,frame,maxFrames)

    progress.total.bar:updateProperties { progress=0, text="" }
    progress.frame.bar:updateProperties { progress=0, text="" }

    -- Set the camera position, target, and up vectors
    Render.camNode:setPinValue(octane.P_TARGET,   {t[1],t[2],t[3]})
    Render.camNode:setPinValue(octane.P_POSITION, {p[1],p[2],p[3]})
    Render.camNode:setPinValue(octane.P_UP,       {u[1],u[2],u[3]})

    Render.is_cancelled = false
    Render.check_quick_render = true

    local kernelNode = Render.targetNode:getInputNode(octane.P_KERNEL)
    kernelNode:setPinValue(octane.P_MAX_SAMPLES, keyframe.samples.slider.value, true)

    if solar.enableSun.check.checked and Render.sunDirNode ~= nil then
        -- overrideTime is true only if the 'Preview Time' slider is being moved
        -- within the solar tab meaning that a time has already been set for previewing.
        if solar.overrideTime == false then
            Render:setupSolarAnimator(maxFrames)
            solar.frameTime = solar.startTime.slider.value + ((frame-1) * solar.timeDiffPerFrame)
            Render.sunDirNode:setPinValue(octane.P_HOUR, solar.frameTime)
        end
    end    

    octane.render.start
    {
        renderTargetNode = Render.targetNode,
        callback         = Render.callback
    }

    if Render.is_dead then 
         return true
    end

    kernelNode:setPinValue(octane.P_MAX_SAMPLES, Render.maxSamples, true)

    progress.total.bar:updateProperties  { progress=0, text="Finished" }
    progress.frame.bar:updateProperties  { progress=0, text="" }
end

--------------------------------------------------------------------------------
-- Locate and calculate the interpolated values for the current showFrame
-- slider value and render out the frame.
--------------------------------------------------------------------------------
function keyframe:renderFrame()

    if keyframe.rebuildOrcCache then
        keyframe:generateOrcCache()
    end

    frame = keyframe.showFrame.slider.value
    if frame >#keyframe.orcCache then
        frame = #keyframe.orcCache
    end

    local t, p, u = keyframe:unpackFrameData(frame)

    octane.render.callbackStop()
    Render.is_keyframe = true
    renderKeyFrame(t,p,u, frame, keyframe.showFrame.slider.maxValue)

    Render.is_keyframe = false

    return true
end
