
-- :::::::::::::::::::::::::::::
-- ::                         ::
-- ::      Project Tab        ::
-- ::                         ::
-- :::::::::::::::::::::::::::::................................................

project  = widget:create()
project:new('button','path')        {"Load ", tooltip="Load a saved project"}:append('editor') {width=320}
project:new('button','saveProject') {"Save", tooltip="Save current data and settings as a project."}

project.changed = false -- Used to determine when essential items are
                        -- changed so that user may be prompted for
                        -- saving a changed project.
 
-- --------------------------------------------------------------------
-- Solar settings are saved by project and so its tab is included here.
-- --------------------------------------------------------------------
solarTab = widget:new('tabs')
      { 
          headers  = {  "Solar Settings" },
          children = {  solar.group }
      }

--------------------------------------------------------------------------------
-- Screen groups and layout for the Settings tab
--------------------------------------------------------------------------------
do 
    local children =
    {
        {project.path.button, project.path.editor, project.saveProject.button }
    }
    local g1 = project:new('group') { children, text=" Project ",
                                       border=true, padding={10}, x=8 }

    ---------------------------------------------------------------------
    children =
    {
        { g1.group },
        { solarTab.tabs },
    }
    project:append('group') { children, padding={0,10}, center=true }
end 

--------------------------------------------------------------------------------
-- Project functions ///////////////////////////////////////////////////////////
--------------------------------------------------------------------------------
function project:enable()
 
    local enabler = cameraFile.editor.text ~= ""

    self.saveProject.button:updateProperties  { enable=enabler }
    self.path.button:updateProperties         { enable=true }
end

--------------------------------------------------------------------
function project:disable()

    self.saveProject.button:updateProperties  { enable=false }
    self.path.button:updateProperties         { enable=false }
end

--------------------------------------------------------------------------------
-- Save common variables needed for restoring work to the current state.
-- This basically means saving the frame totals for keyframes, the name of
-- the Orc file involved and any daylight environment that may be enabled.
--------------------------------------------------------------------------------
local function saveProject(name)

    Render:cancel()

    local file, status = io.open(name, "wb")
    if status then
        local msg = "Unable to create "..name.."\nMake sure directory exists.\n"
        msg = msg:gsub("\\", "/")
        msg = msg.."Status:"..status
        dialog.critical(msg, false)
        return false
    end

    if cameraFile.editor.text ~= "" then
        local path,filename,ext = names.validate:parsePath(name)
        cameraFile.editor.text = "Project: "..filename
    end

    local endL = "\n"
    file:write("OrcVersion="                    .. VERSION..endL)
    file:write("cameraFile.editor.text='"       .. cameraFile.editor.text.."'\n")
    file:write("keepCamera.check.checked="      .. tostring(keepCamera.check.checked)..endL)
    file:write("renderTarget.combo.selectedIx=" .. renderTarget.combo.selectedIx..endL)

    file:write("animate.path.path='"                   .. animate.path.path .. "'\n")
    file:write("animate.path.editor.text='"            .. animate.path.editor.text .. "'\n")
    file:write("animate.imageFormat.combo.selectedIx=" .. animate.imageFormat.combo.selectedIx..endL)
    file:write("animate.videoFormat.combo.selectedIx=" .. animate.videoFormat.combo.selectedIx..endL)
    file:write("animate.samples.slider.value="         .. animate.samples.slider.value..endL)
    file:write("animate.frameRate.slider.value="       .. animate.frameRate.slider.value..endL)
    file:write("animate.shutterTime.slider.value="     .. animate.shutterTime.slider.value..endL)
    file:write("animate.forceSequence.check.checked="  .. tostring(animate.forceSequence.check.checked)..endL)
    file:write("animate.runFFmpeg.check.checked="      .. tostring(animate.runFFmpeg.check.checked)..endL)
    file:write("animate.startFrame.slider.value="      .. animate.startFrame.slider.value..endL)
    file:write("animate.endFrame.slider.value="        .. animate.endFrame.slider.value..endL)
    file:write("animate.step.slider.value="            .. animate.step.slider.value..endL)
    file:write("animate.startSequence.numeric.value="  .. animate.startSequence.numeric.value..endL)

    file:write("keyframe.smoothing.numeric.value="     .. keyframe.smoothing.numeric.value..endL)
    file:write("keyframe.shutterTime.slider.value="    .. keyframe.shutterTime.slider.value..endL)

    file:write("solar.enableSun.check.checked="  .. tostring(solar.enableSun.check.checked)..endL)
    file:write("solar.startTime.slider.value="   .. solar.startTime.slider.value ..endL)
    file:write("solar.endTime.slider.value="     .. solar.endTime.slider.value ..endL)
    file:write("solar.preview.slider.value="     .. solar.preview.slider.value ..endL)
    file:write("solar.startHMS.label.text='"     .. solar.startHMS.label.text .."'\n")
    file:write("solar.endHMS.label.text='"       .. solar.endHMS.label.text .."'\n")
    file:write("solar.previewHMS.label.text='"   .. solar.previewHMS.label.text .."'\n")

    if solar.localTime ~= nil then
        file:write("solar.localTime=" .. solar.localTime ..endL)
    end

    names:saveDefaults(file)

    if cameraFile.data ~= nil then
        ---------------------------------
        local function vectorToString(v)
            return string.format("%s,%s,%s", v[1], v[2], v[3])
        end
        ---------------------------------

        if cameraFile.totalFrames > 1 then
            file:write("cameraFile.data={\n")
            for i=1, cameraFile.totalFrames do
                local t, p, u = cameraFile:unpackFrameData(i)
                local s = string.format("{%s,%s,%s},\n", vectorToString(t),vectorToString(p),vectorToString(u))
                file:write(s)
            end
            file:write("}\n")
        end
    end

    ----------------------------------------------------
    -- Write out the keyframe.item table in blocks of 20
    ----------------------------------------------------
    file:write("kfItem = {" .. endL)
    local c = 0
    local buff = ""
    for i,v in pairs(keyframe.item) do
        c = c + 1
        buff = buff .. v .. ","
        if c == 20 then
            file:write(buff .. endL)
            buff = ""
            c = 0
        end
    end
    file:write(buff .. "}" .. endL)
    -----------------------------------

    file:flush()
    file:close()

    return true
end

--------------------------------------------------------------------------------
-- Loadng a saved project runs code previously saved as an .orcs file
-- in the saveProject function. 
--------------------------------------------------------------------------------
local function loadProject(name)

    local status, fullname, filename = names.validate:pathForOpen(project.path.editor.text, "orcs")
    if status == true then
        project.path.editor.text = fullname
    else
        return false
    end

    keyframe.rebuildOrcCache = true
    local ff, status = loadfile(fullname)

    if status then
        local msg = "Unable to load project file: "..filename.."\n"
        msg = msg:gsub("\\", "/")
        msg = msg.."Status:"..status
        dialog.critical(msg, false)
        return false
    else
        local st = solar.localTime
        solar.localTime = nil
        status = ff()
        if status then
            local msg = "Unable to execute project file: "..filename.."\n"
            msg = msg:gsub("\\", "/")
            msg = msg.."Status:"..status
            dialog.critical(msg, false)
            if st ~= nil then
                solar.localTime = st
            end
            return false
        end

        cameraFile.totalFrames = #cameraFile.data
        cameraFile.vectors = 3

        -- Check needs to be made for saved combo item as the number of render targets
        -- may have changed leaving it hanging. Default to 1 in that case.
        if renderTarget.combo.selectedIx > #Render.targets then
            renderTarget.combo.selectedIx = 1
        end
        keyframe.samples.slider.value = 1
        --animate.samples.slider.value = animate.samples.slider.maxValue
    end
    return true
end

--------------------------------------------------------------------
local function showSuccessfulSave(filename, proceedType)
    local msg1 = "The project file:"
    local msg2 = filename.."\n\n"
    msg2 = msg2 .. "has was successfully saved.\n"
    project.changed = false
    dialog.info(msg1, msg2)
end

--------------------------------------------------------------------
-- Checking for essential changes gives the user a chance to save
-- those changes as a project.
--------------------------------------------------------------------
function project:checkForChange(proceedType)

    Render:cancel()
    if self.changed == false then
        return "continue"
    end

    local path,filename,ext = names.validate:parsePath(self.path.editor.text)

    if self.path.editor.text == "" then
        local answer = dialog.question(3, "Changes have been made.",
                          "Would you like to save as a new project?")
        if answer == 1 then
            if self.saveProject.button.callback() == true then
               showSuccessfulSave(self.path.editor.text, proceedType)
               return "saved"
            end
        elseif answer == 3 then
            return "cancel"
        end
        return "continue"
    end

    local msg = filename.." has changed."
    local answer = dialog.question(3, msg, "Save changes to the current project?")
    if answer == 1 then
        gui:enable()
        if saveProject(self.path.editor.text) == false then
            self.changed = false
            return "continue"
        end
        showSuccessfulSave(filename, proceedType)
        return "saved"
    elseif answer == 2 then
        return "continue"
    end
    return "cancel"
end

-------------------------------------------------------------------
function project:clearProjectChanged()
    animate.imageFormat.ignoreCallback = true
    animate.videoFormat.ignoreCallback = true
    self.changed = false
end

-------------------------------------------------------------------
-- Settings gui callbacks
-------------------------------------------------------------------
function project:setupCallbacks()

    ---------------------------------------------------------------
    project.saveProject.button.callback = function()

        -- choose an output file for saving project
        Render:cancel()
        local outputName = ""
        if project.path.editor.text ~= "" then
            outputName = project.path.editor.text
        else
            outputName = Render.projectPath
        end
        local ret = octane.gui.showDialog
        {
            type      = octane.gui.dialogType.FILE_DIALOG,
            title     = "Choose the output name for project",
            path      = outputName,
            wildcards = "*.orcs",
            save      = true
        }

        -- if a file is chosen
        if ret.result ~= "" then
            local name = ret.result:gsub("\\", "/")
            local status, fullname, filename, ext = names.validate:pathForSave(name, "orcs")
            if status == false then
                return false
            end

            if ext == "" then  -- if ext is blank, no check was made for existence
                local file = io.open(fullname)
                if file ~= nil then
                    file:close()
                    local msg = filename.."\nalready exists."
                    local answer = dialog.question(2, msg, "Do you want to replace it?")
                    if answer ~= 1 then
                        return false
                    end
                end
            end

            project.path.editor.text = fullname
            saveProject(fullname)
            gui:enable()
            project:clearProjectChanged()
            return true
        end

        return false
    end

    ---------------------------------------------------------------
    project.path.button.callback = function()

        Render:cancel()
        local answer = project:checkForChange("project")
        if answer == "cancel" then
            return false
        end

        keyframe.showFrame.slider.value = 1

        -- choose a project file
        local ret = octane.gui.showDialog
        {
            type      = octane.gui.dialogType.FILE_DIALOG,
            title     = "Choose a project file",
            path      = Render.projectPath,
            wildcards = "*.orcs",
            save      = false,
        }

        -- if a file is chosen
        if ret.result ~= "" then
            project.path.editor.text = ret.result:gsub("\\", "/")
            if not loadProject(project.path.editor.text) then
                project.path.editor.text = ""
                project:clearProjectChanged()
                return false
            end
            if cameraFile.editor.text ~= "" then
                -- Preserve start and end frame so initializing won't lose them.
                local startFrame = animate.startFrame.slider.value
                local endFrame = animate.endFrame.slider.value

                gui:enable()
                keyframe:initializeData()
                keyframe.totalFrames = 0  -- Total number of interpolated frames
                for i,v in pairs(kfItem) do
                    if keyframe.item[i] ~= nil then
                        keyframe.item[i] = v
                        if i ~= #kfItem then
                            keyframe.totalFrames = keyframe.totalFrames + v
                        end
                    else
                        break
                    end
                end
                -- maxValue should always show the total number of interpolated frames
                keyframe.showFrame.slider.maxValue = keyframe.totalFrames
                animate.startFrame.slider.maxValue = keyframe.totalFrames
                animate.endFrame.slider.maxValue   = keyframe.totalFrames
                -- Restore project endFrame
                animate.startFrame.slider.value = startFrame
                animate.endFrame.slider.value = endFrame
                if animate.startFrame.slider.value >  animate.startFrame.slider.maxValue then
                    animate.startFrame.slider.value = animate.startFrame.slider.maxValue
                end
                if animate.endFrame.slider.value >  animate.endFrame.slider.maxValue then
                    animate.endFrame.slider.value = animate.endFrame.slider.maxValue
                end

                keyframe:loadScreenItems()
                keyframe:renderFrame()
                solar:enable {check=true}
                project:clearProjectChanged()
                return true
            end
        else
            project:clearProjectChanged()
            return false
        end

        gui:disable()
        cameraFile.editor.text = "" 
        cameraFile.open.button.enable = true
        project.path.button.enable = true
        project:clearProjectChanged()
    end

    ---------------------------------------------------------------
    return true
end

