
-- ::::::::::::::::::::::::::::::::::
-- ::                              ::
-- ::      Camera file  section    ::
-- ::                              ::
-- ::::::::::::::::::::::::::::::::::...........................................

cameraFile        = widget:create():append('editor') {width=320}
cameraFile.open   = cameraFile:new('button') {"Open",   tooltip="Select an ORC file"}
cameraFile.reload = cameraFile:new('button') {"Reload", tooltip="Reload the current ORC file" }

do ----------------------------------------------------------------------

    local children =  -- Screen group objects are in row and column order.
    {
        { cameraFile.open.button, cameraFile.editor, cameraFile.reload.button }
    }
    cameraFile:append('group') { children, text=" Camera File ",
                                 border=true, padding={10}, x=8 }
end

--------------------------------------------------------------------------------
-- cameraFile functions ////////////////////////////////////////////////////////
--------------------------------------------------------------------------------
function cameraFile:enable()
 
    if cameraFile.editor.text ~= "" then
        local embedCheck = string.match(cameraFile.editor.text, "Project: ")
        if embedCheck ~= nil then
            cameraFile.reload.button:updateProperties  { enable=false }
        else
            cameraFile.reload.button:updateProperties  { enable=true }
        end
    end

    cameraFile.open.button:updateProperties { enable=true }
end

--------------------------------------------------------------------------------
function cameraFile:disable()

    cameraFile.open.button:updateProperties   { enable=false }
    cameraFile.reload.button:updateProperties { enable=false }
end

--------------------------------------------------------------------------------
function cameraFile:unpackFrameData(frame)

    local c = self.data[frame]

    local t = {c[1],c[2],c[3]}
    local p = {c[4],c[5],c[6]}
    local u = nil
    if self.vectors == 3 then
        u = {c[7],c[8],c[9]}
    else
        u = {0,1,0}
    end
    return t, p, u
end

--------------------------------------------------------------------------------
-- Read in the data from the camera file. Files may consist of lines 
-- containing 2 vectors (6 fields) or 3 vectors (9 fields)
-- Two vector files contain target and position vectors while 3 vector files
-- also have an Up vector.
--------------------------------------------------------------------------------
function cameraFile:load()

    project:clearProjectChanged()

    local status, fullname, filename = names.validate:pathForOpen(cameraFile.editor.text, "orc")
    if status == true then
        cameraFile.editor.text = fullname
    else
        return false
    end

    cameraFile.data = {}
    cameraFile.totalFrames = 0

    local file, status = io.open(cameraFile.editor.text, "r")

    if status then
        local msg = "Couldn't open camera file: " .. filename .. "\n"
        msg = msg .. "Status: " .. status
        dialog.critical(msg, false)
        return false
    end

    local vtype = file:read("*l")
    local totalFields = 0
    if vtype == nil then
        local msg1 = "The Orc file:"
        local msg2 = cameraFile.editor.text .. "\n\nappears to be empty.\n"
        msg2 = msg2.."Try adding frames with Alt A or recreate it in exporter."
        dialog.info(msg1, msg2)
        file:close()
        return false
    end

    for vert in string.gmatch(vtype, "([^,]+)") do totalFields = totalFields+1 end
    if totalFields == 9 then
        cameraFile.vectors = 3
    elseif totalFields == 6 then
        cameraFile.vectors = 2
    else
        local msg1 = "The Orc file:"
        local msg2 = cameraFile.editor.text.."\n\nmust consist of 6 or 9 numeric fields per line.\n"
        msg2 = msg2.."These should be floating point numbers separated by comas."
        dialog.info(msg1, msg2)
        file:close()
        return false
    end

    local lineCount = 0
    file:seek("set",0)
    for sline in file:lines() do 
        local i = 0
        local vectorLine = {}
        for vert in string.gmatch(sline, "([^,]+)") do
            i = i+1
            vectorLine[i] = vert
        end

        lineCount = lineCount + 1
        if i ~= totalFields then
            local msg1 = "The Orc file:"
            local msg2 = cameraFile.editor.text.."\n\nhas and error at line "..lineCount.."\n"
            msg2 = msg2.."Was expecting "..totalFields.." fields but got "..i..".\n"
            msg2 = msg2.."Would you like to ignore this error and continue?\n"
            local answer = dialog.question(2, msg1, msg2)
            if answer == 2 then
                file:close()
                return false
            end
        else
            table.insert( cameraFile.data, vectorLine )
        end
    end

    file:close()

    -- Because key framing requires two points, make sure there are two data records.
    if #cameraFile.data == 1 then
        table.insert( cameraFile.data, cameraFile.data[1])
    end
    cameraFile.totalFrames = #cameraFile.data

    return true
end


--------------------------------------------------------------------------------
-- Establish the callbacks for cameraFile components
--------------------------------------------------------------------------------
function cameraFile:setupCallbacks()

    ---------------------------------------------------------------
    cameraFile.open.button.callback = function()
        local answer = project:checkForChange("Orc")
        if answer == "cancel" or answer == "saved" then
            return false
        end
        -- choose an input file
        local ret = octane.gui.showDialog
        {
            type      = octane.gui.dialogType.FILE_DIALOG,
            title     = "Choose a camera file",
            path      = Render.projectPath,
            wildcards = "*.orc",
            save      = false,
        }

        -- if a file is chosen
        if ret.result ~= "" then
            project.path.editor.text = ""
            animate.path.editor.text = ""
            animate.path.editor.path = ""
            cameraFile.editor:updateProperties{ text=ret.result:gsub("\\", "/") }
            if cameraFile:load() then
                gui:enable()
                keyframe:initializeData()
                keyframe:renderFrame()
                return true
            end
        else
            return false
        end

        gui:disable()
        cameraFile.editor.text = ""
        -- If no cameraFile is available, disable gui except for the open buttons
        -- for the cameraFile or project.
        cameraFile.open.button.enable = true
        project.path.button.enable = true

        return true
    end
    ---------------------------------------------------------------

    cameraFile.reload.button.callback = function()
        local answer = project:checkForChange("Orc")
        if answer == "cancel" or answer == "saved" then
            return false
        end
        if cameraFile:load() then
            keyframe:initializeData()
            keyframe:renderFrame()
        end
    end

    ---------------------------------------------------------------
end
