
-- ----------------------------------
--  Create all the ORC gui elements 
-- ----------------------------------
require "Orc_bin/widget"
require "Orc_bin/Tab/names"
require "Orc_bin/Tab/settings"
require "Orc_bin/Tab/solar"
require "Orc_bin/Tab/project"
require "Orc_bin/Tab/keyframe"
require "Orc_bin/Tab/animate"
require "Orc_bin/cameraFile"

-- ::::::::::::::::::::::::::::::::
-- ::                            :: 
-- ::   Render Target  section   ::  renderTarget is filled in later on,
-- ::                            ::  so just make its container.
-- ::::::::::::::::::::::::::::::::.............................................

renderTarget = widget:create()
renderTarget:append('combo') { label={"Render Target",width=105},
                               items={""}, width=140
                             } 
renderTarget.ignoreCallback = false -- Need to ignore the callback at times

exit       = widget:new('button') {"Exit"}
keepCamera = widget:new('check')  { text="Retain camera position on Exit", width=182}

--------------------------------------------------------------------------------------
-- Screen groups and layout for non-tabbed gui components
--------------------------------------------------------------------------------------
do
    local children =
    {
        { renderTarget.label, renderTarget.combo }
    }
    renderTarget:append('group') { children, text="Select render target" }

    ---------------------------------------------------------------------
    children =
    {
        { keepCamera.check }
    }
    keepCamera:append('group') { children }

    ---------------------------------------------------------------------
    children =
    {
        { renderTarget.group, keepCamera.group, exit.button }
    }
    rt_keepCamera = widget:new('group') { children, padding={7,10} }

end ----------------------------------------------------------------------------

-- :::::::::::::::::::::::::::::
-- ::                         ::
-- ::      Progress Bar       ::
-- ::                         ::
-- :::::::::::::::::::::::::::::................................................

progress = widget:create()

progress:new('progressBar','total') { text="Total render progress",
                                      width=(cameraFile.group.width * .97)
                                    }

progress:new('progressBar','frame') { text="Current frame progress",
                                      width=(cameraFile.group.width * .97)
                                    }
do  ----------------------------------------------------
    local children =
    {
        { progress.total.bar },
        { progress.frame.bar }
    }
    progress:new('group','bar') { children, border=true, padding={2},inset={10}}
end --::::::::::::::::::::::::::................................................

-- :::::::::::::::::::::::::::::
-- ::                         ::
-- ::      Defaults Tab       ::
-- ::                         ::
-- :::::::::::::::::::::::::::::................................................

defaultTab = widget:new('tabs')
{ 
    headers  = {  "File names", "General Settings" },
    children = {  names.group,  settings.group }
}
-- :::::::::::::::::::::::::::::
-- ::                         ::
-- ::     Main Tab group      ::
-- ::                         ::
-- :::::::::::::::::::::::::::::................................................

local tabGroup = widget:new('tabs')
{ 
    headers  = {  "Key Frame", "Defaults", "Project", "Animation" },
    children = {  keyframe.group, defaultTab.tabs, project.group, animate.group }
}

VisibleTab = tabGroup.tabs  -- Use for checking current tab that is visible.
                            -- VisibleTab.currentTab will be 1, 2, 3, or 4
                            --
                            -------------------------------------------------
tabGroup.tabs.callback = function()    -- When the animation tab is selected,
    if VisibleTab.currentTab == 4 then -- rebuild the output name as changes
                                       -- may have been made in names tab.
        if animate.path.verify == false then -- Only if path is not being edited.
            animate:rebuildOutputName()
        end
    end
end

--/////////////////////////////////////////////////////////////////////////////
--//                                                                         // 
--//               Assemble all groups into the final screen layout          // 
--//                                                                         // 
--/////////////////////////////////////////////////////////////////////////////
do 
    local children =
    {
        { cameraFile.group },
        { rt_keepCamera.group },
        { tabGroup.tabs },
        { progress.bar.group }
    }
    layout = widget:new('group') { children }
end 

--------------------------------------------------------------------------------
gui = {}
--------------------------------------------------------------------------------
function gui:enable(...)
 
    local argv = {...} if (#argv > 0) then argv = argv[1] end

    if renderTarget.combo.enable == false then
        renderTarget.combo.enable = true
    end

    -- Check if the animate path button should be forced on
    if argv.animate_pathOn ~= nil then
        animate:enable {pathOn=true}
    else
        animate:enable()
    end

    local enabler = cameraFile.editor.text ~= ""
    keepCamera.check:updateProperties { enable=enabler }

    cameraFile:enable()
    keyframe:enable()
    solar:enable { check=enabler }
    settings:enable()
    project:enable()
    names:enable()

    scene:getMaxSamples()
end

--------------------------------------------------------------------------------
function gui:disable(...)

    local argv = {...} if (#argv > 0) then argv = argv[1] end
    if argv.all ~= nil then
        renderTarget.combo:updateProperties { enable=false }
    end
    if argv.exceptAnimate ~= nil then
        animate:disable {pathOn=true}
    else
        animate:disable()
    end
    keepCamera.check:updateProperties { enable=false }

    cameraFile:disable()
    keyframe:disable()
    solar:disable()
    settings:disable()
    project:disable()
    names:disable()
end

--------------------------------------------------------------------------------
-- Establish the callbacks to all the GUI elements
--------------------------------------------------------------------------------
function gui:setupCallbacks()

    OrcWindow.callback = function(event)
        if Render.is_running then
            scene:checkForKeepingCameraPosition()
        end
        Render:cancel()

        --  If the window is closed using the OS's X (close) button while
        --  a render is in progress instead of using Orc's Exit button, the
        --  script continues to run. If killed, an error would be
        --  produced without the use of the is_dead flag.
        if Render.is_dead == nil then
            Render.is_dead = false
        else
            Render.is_dead = true
        end
        Render.is_cancelled = true
    end

    cameraFile:setupCallbacks()
    keyframe:setupCallbacks()
    solar:setupCallbacks()
    animate:setupCallbacks()
    settings:setupCallbacks()
    project:setupCallbacks()
    names:setupCallbacks()

    ---------------------------------------------------------------

    renderTarget.combo.callback = function()
        -- Reset the ignoreCallback flag if set and return
        if renderTarget.ignoreCallback == true then
            renderTarget.ignoreCallback = false
            return false
        end
        if renderTarget.combo.enable == false then
            return false
        end

        scene:getCopy()
        if solar.enableSun.check.checked then
            solar.enableSun.check.callback()
        end
        names:buildNames(true)
        if Render.is_running then
            scene:getMaxSamples()
            Render.reset_animator = true

            if keyframe.activeButton ~= nil then
                gui:disable()
                keyframe.activeButton:updateProperties {enable=true}
            end
        else
            if Window_is_showing then
                gui:enable() 
                if cameraFile.editor.text ~= "" then
                    Render:cancel()
                    keyframe:renderFrame()
                end
            end
        end
    end

    ---------------------------------------------------------------

    exit.button.callback = function()
        Render:cancel()
        local answer = project:checkForChange("Orc")
        if answer == "cancel" then
            return false
        end
        scene:checkForKeepingCameraPosition()
        OrcWindow:closeWindow()
    end
end
