------------------------------------------------------------------------------------------
-- Globals for handling various aspects of rendering.
------------------------------------------------------------------------------------------
Render = {
    scriptsPath  = octane.file.getSpecialDirectories()["userScriptDirectory"],
    projectPath  = octane.file.getParentDirectory(octane.project.getCurrentProject()),
    targets      = octane.project.getSceneGraph():findNodes(octane.NT_RENDERTARGET, true),
    check_quick_render = false,  -- Used to quiet progress bar on short renders
    is_keyframe  = false,        -- Render is for keyframe preview mode.
    is_cancelled = false,        -- Render has been cancelled.
    is_running   = false,        -- A render is in progress
    is_dead      = nil,          -- Tries to handle window close other than exit button
    maxSamples   = 1,            -- Maximum samples for this render
    sceneNode    = nil,          -- copy of scene's root node
    targetNode   = nil,          -- Render target node for this render
    camNode      = nil,          -- Camera node for this render
    envNode      = nil,          -- Environment node for this render
    sunDirNode   = nil,          -- Sun direction node for this render
    -- The following are used for constructing file names of output renders
    -- so as to easily identify the many image files that are possible.
    samplesPerPixel = 1,
    resolution  = {              -- Output resolution of image
        node = nil,
        x = 0,
        y = 0
    },
    kernel = {
        node   = nil,            -- Kernel node
        kType  = ""              -- Kernal type: DL, PT, PMC, INFO, MP
    },
    infoChannel = {              -- InfoChannel when kernelType is INFO
        node    = nil,          
        atValue = 0,             -- enum value of infoChannel type           
        icType  = ""             -- "GEOMETRIC_NORMALS", "WIREFRAME", etc. 
    }
}


-- Octane 2 compatibility hack
if not octane.imageSaveType then octane.imageSaveType = octane.render.imageType end

-----------------------------------------------------------------------------
-- Get the samples per pixel for the current render target and kernel.
-----------------------------------------------------------------------------
function Render:getSamplesPerPixel()

    self.samplesPerPixel = 1
    if self.targetNode ~= nil then
        local kernelNode = self.targetNode:getInputNode(octane.P_KERNEL)
        self.samplesPerPixel = kernelNode:getPinValue(octane.P_MAX_SAMPLES)
    end
end

-----------------------------------------------------------------------------
function Render:getResolution()

    self.resolution = {}
    self.resolution.node = nil
    self.resolution.x = 0
    self.resolution.y = 0

    if self.targetNode ~= nil then
        self.resolution.node = self.targetNode:getInputNode(octane.P_RESOLUTION)
        local values = self.resolution.node:getAttribute(octane.A_VALUE)
        if values ~= nil then
            self.resolution.x = values[1]
            self.resolution.y = values[2]
            --values[1] = 1200
            --values[2] = 1200
            --self.resolution.node:setAttribute(octane.A_VALUE, values)
        end
    end
end

-----------------------------------------------------------------------------
function Render:getKernelType()

    self.kernel.node = nil
    self.kernel.kType = ""

    if self.targetNode ~= nil then
        local kernelNode = self.targetNode:getInputNode(octane.P_KERNEL)
        local kt = nil
        if kernelNode.type == octane.NT_KERN_DIRECTLIGHTING then kt = "DL" end
        if kernelNode.type == octane.NT_KERN_PATHTRACING then kt = "PT" end
        if kernelNode.type == octane.NT_KERN_PMC then kt = "PMC" end
        if kernelNode.type == octane.NT_KERN_INFO then kt = "INFO" end
        if kernelNode.type == octane.NT_KERN_MATPREVIEW then kt = "MP" end
        self.kernel.node  = kernelNode
        self.kernel.kType = kt
    end
end

-----------------------------------------------------------------------------
function Render:getInfoChannelType()

    self.infoChannel = {}
    self.infoChannel.node = nil
    self.infoChannel.atValue = 0
    self.infoChannel.icType = ""

    if self.targetNode ~= nil then
        local kernelNode = self.targetNode:getInputNode(octane.P_KERNEL)
        if kernelNode.type == octane.NT_KERN_INFO then
            self.kernelType = "INFO"
            self.infoChannel = {}
            self.infoChannel.node = kernelNode:getInputNode(octane.P_INFOCHANNELS_TYPE)
            --[[
            print("InfoNode:", self.infoChannel.node)
            info = self.infoChannel.node:getNodeInfo()
            print("---------Info-----------")
            for i,v in pairs(info) do print(i,v) end
            print("-------Info End--------")
            --infoProps = self.infoChannel.node:getProperties()
            --for i,v in pairs(infoProps) do print(i,v) end
            ]]
            atValue = self.infoChannel.node:getAttribute(octane.A_VALUE)
            local itype = -1
            if atValue == 0 then
                 itype = "gn"          -- Geometric Normals
            elseif atValue == 1 then
                 itype = "sn"          -- Shading Normals
            elseif atValue == 2 then
                 itype = "pos"         -- Position
            elseif atValue == 3 then
                 itype = "zd"          -- Z depth
            elseif atValue == 4 then
                 itype = "mId"         -- Material ID
            elseif atValue == 5 then
                 itype = "tx"          -- Texture coordinates
            elseif atValue == 7 then
                 itype = "wf"          -- Wireframe
            elseif atValue == 8 then
                 itype = "ivn"         -- Interpolated vertex normals
            elseif atValue == 9 then
                 itype = "obId"        -- Object layer ID
            elseif atValue == 10 then
                 itype = "AO"          -- Ambient Occlusion
            else
                 itype = ""            -- Unknown
            end
            self.infoChannel.atValue = atValue
            self.infoChannel.icType = itype
        end
    end
end

--------------------------------------------------------------------------------
--       Update statistics during render process through callback
--------------------------------------------------------------------------------
function Render.callback(rc)

    if Render.is_cancelled then
        return
    end
    
    -- Get maximum samples to render for this frame
    --
    local ms  = Render.is_keyframe and keyframe.samples.slider.value
               or animate.samples.slider.value

    local tps = rc.renderTime / rc.samples  -- render time per sample
    local pc = rc.samples / ms    -- percent complete for this frame
    local rs = ms - rc.samples    -- remaining samples for this frame
    local et = tps * rs           -- remaining time for this frame

    -- On the very first callback for rendering a frame, only show
    -- progress if there are still more than 3 seconds to go.
    --
    if Render.check_quick_render then
        if et < 3.1 then
            progress.frame.bar:updateProperties { progress=0, text="" }
            return
        end
        Render.check_quick_render = false
    end

    local h = math.floor(et/3600)
    local m = math.floor((et-h*3600)/60)
    local s = math.floor(et-h*3600-m*60)

    progress.frame.bar:updateProperties {
        progress=pc, text=string.format("%.1f%% - %02d:%02d:%02d",100*pc,h,m,s)
    }
end

--------------------------------------------------------------------------------
function Render:cancel()

    self.is_cancelled = true
    octane.render.callbackStop()
    octane.render.clear()
    collectgarbage()
end

----------------------------------------------------------------------------------
--                Calculate values for solar animation
--------------------------------------------------------------------------------
function Render:setupSolarAnimator(totalFrames)

    if self.sunDirNode ~= nil then
        solar.frameTime = solar.startTime.slider.value
        local totalTime = solar.endTime.slider.value - solar.startTime.slider.value
        solar.timeDiffPerFrame = totalTime / (totalFrames-1)
    else
        solar.frameTime = solar.startTime.slider.value
        solar.timeDiffPerFrame = 0
    end
end

--------------------------------------------------------------------------------
--     Use Octane's animator function so motion blur can be shown
--------------------------------------------------------------------------------
function Render:setupCameraAnimator(startFrame, endFrame, cStep)

    local target = {}
    local position = {}
    local up = {}
    local v = {}
    
    -- We've swapped nFrame with endFrame so that  all interpolated values
    -- are placed in the animator. This is needed to match motion blur
    -- when rendering all frames or just a range of frames due to the
    -- timing that is necessary.
    local nFrame = keyframe.showFrame.slider.maxValue
    for frame = startFrame, nFrame, cStep do
        if frame > #keyframe.orcCache then
            break
        end
        local t, p, u = keyframe:unpackFrameData(frame)
        table.insert(target,   t)
        table.insert(position, p)
        table.insert(up,       u)
    end

    local total = #target - 1
    local shutterTime = self.is_keyframe and keyframe.shutterTime.slider.value
                        or animate.shutterTime.slider.value
    if octane.render.setShutterTime then
        -- older versions of Octane
        octane.render.setShutterTime(shutterTime)
    else
        local animationSettingsNode = self.targetNode:getInputNode(octane.P_ANIMATION)
        animationSettingsNode:setPinValue(octane.P_SHUTTER_TIME, shutterTime)
    end
    self.camNode:getInputNode(octane.P_TARGET)
          :setAnimator(octane.A_VALUE, {0}, target, 1/total)

    self.camNode:getInputNode(octane.P_POSITION)
          :setAnimator(octane.A_VALUE, {0}, position, 1/total)

    self.camNode:getInputNode(octane.P_UP)
          :setAnimator(octane.A_VALUE, {0}, up, 1/total)

    self:setupSolarAnimator(keyframe.showFrame.slider.maxValue)
end

---------------------------------------------------------------
--  Clear the camera animator
---------------------------------------------------------------
local function clearCameraAnimator()
    Render.camNode:getInputNode(octane.P_TARGET):clearAnimator(octane.A_VALUE)
    Render.camNode:getInputNode(octane.P_POSITION):clearAnimator(octane.A_VALUE)
    Render.camNode:getInputNode(octane.P_UP):clearAnimator(octane.A_VALUE)
end

---------------------------------------------------------------
--  Clear the sun animator and restore the local time.
---------------------------------------------------------------
local function clearSolarAnimator()

    if solar.enableSun.check.checked and Render.sunDirNode ~= nil then
        Render.sunDirNode:setPinValue(octane.P_HOUR, solar.startTime.slider.value)
    end

end

--------------------------------------------------------------------------------
local function clearAnimators()
    clearCameraAnimator()
    clearSolarAnimator()
end

------------------------------------------------------------------------------
--           Save the current frame to the animation output path
------------------------------------------------------------------------------
local function saveFrameImage(path, frame)

    local function getPath(path, frame)
        
        local sequenceMatch = "000000"
        local seqPattern = "%0"..sequenceMatch:len().."d"
        local suffix = animate:get('imageExtension')
        local file = path .. seqPattern:format(frame) .. suffix
        return file
    end

    local out = getPath(path, frame)
    if octane.render.saveImage(out, animate:get('imageFormat')) == false then
        local msg = "Unable to create "..out.."\nMake sure directory exists.\nCanceling render."
        msg = msg:gsub("\\", "/")
        dialog.critical(msg, false)
    end
end

--------------------------------------------------------------------------------
--
--                      Begin rendering a range of frames 
--
--------------------------------------------------------------------------------
function Render:start(path, startFrame, endFrame, step, saveImage)

    local shutterTime = self.is_keyframe and keyframe.shutterTime.slider.value
                        or animate.shutterTime.slider.value
    if octane.render.setShutterTime then
        -- older versions of Octane
        octane.render.setShutterTime(shutterTime)
    else
        local animationSettingsNode = self.targetNode:getInputNode(octane.P_ANIMATION)
        animationSettingsNode:setPinValue(octane.P_SHUTTER_TIME, shutterTime)
    end
        
    self:setupCameraAnimator(startFrame,endFrame, step)
 
    self.is_cancelled = false
    self.reset_animator = false
    self.is_running = true

    local checkFrame = (endFrame-startFrame+1) / step
    local total_frames = math.floor(checkFrame)
    if checkFrame > total_frames then
        total_frames = total_frames + 1
    end

    local currentTime = 0
    local startTime = os.clock()

    -- Show initial frame progress bar
    --
    progress.total.bar:updateProperties
        { progress=0, text=string.format("Rendering frame 1/%d", total_frames) }
    progress.frame.bar:updateProperties { progress=0, text="" }

    -- Set maximum samples in kernel
    --
    local kernelNode = self.targetNode:getInputNode(octane.P_KERNEL)
    local samples = self.is_keyframe and keyframe.samples.slider.value
                    or animate.samples.slider.value
    kernelNode:setPinValue(octane.P_MAX_SAMPLES, samples, true)

    --------------------------------------------------------------------
    --
    --               Cycle through all the frames
    --
    --------------------------------------------------------------------
    local frameID = animate.startSequence.numeric.value
    local count = 1
    for frame = startFrame, endFrame, step do

        -- If the following is true, the render target has been modified.
        -- Adjust accordingly and allow rendering to continue.
        -- self.reset_animator = false
        if self.reset_animator then
            self.reset_animator = false
            clearSolarAnimator()
            self:setupCameraAnimator(startFrame,endFrame, step)
            kernelNode = self.targetNode:getInputNode(octane.P_KERNEL)
            kernelNode:setPinValue(octane.P_MAX_SAMPLES, samples, true)
        end

        -- Set the time in the scene 
        self.sceneNode:updateTime(currentTime)

        -- Keep the keyframe slider value in sync with current frame
        keyframe.showFrame.slider.value = frame

        -- The quick render check will be done with the first callback.
        -- Progress won't be shown for 3 second renders, just because.
        self.check_quick_render = true
 
        -- Set the sun for the right time for this frame
        if solar.enableSun.check.checked and self.sunDirNode ~= nil then
            solar.frameTime = solar.startTime.slider.value + ((frame) * solar.timeDiffPerFrame)
            self.sunDirNode:setPinValue(octane.P_HOUR, solar.frameTime)
        end

        -- --------------------------------------------------------
        --                     Render the Frame
        -- --------------------------------------------------------
        octane.render.start
        {
            renderTargetNode = self.targetNode,
            callback         = self.callback,
        }
        -- --------------------------------------------------------

        if self.is_dead then break end

        if self.is_cancelled then 
            progress.total.bar:updateProperties { progress=0, text="cancelled" } 
            break 
        end

        -- --------------------------------------------------------
        --                        Save the image
        -- --------------------------------------------------------
        if saveImage then
            saveFrameImage(path, frameID)
            -- Saving may fail causing cancel 
            if self.is_cancelled then
                progress.total.bar:updateProperties { progress=0, text="cancelled" } 
                break 
            end
        end
        if animate.forceSequence.check.checked == true then
            frameID = frameID + 1
        else
            frameID = frameID + step
        end
        -- --------------------------------------------------------
        --          Update the time for the next frame
        -- --------------------------------------------------------
        -- We need to count time for all interpolated frames to
        -- to properly handle motion blur. nFrame has been swapped
        -- with endFrame to make that happen.
        local nFrame = keyframe.showFrame.slider.maxValue
        currentTime = (frame+1-startFrame) * (1 / (nFrame-startFrame))

        -- --------------------------------------------------------
        --               Calculate remaining time
        -- --------------------------------------------------------
        local et = (total_frames-count) * (os.clock()-startTime) / count
        local h = math.floor(et/3600)
        local m = math.floor((et-h*3600)/60)
        local s = et-h*3600-m*60

        -- --------------------------------------------------------
        --                   Update the progress bar
        -- --------------------------------------------------------
        progress.total.bar:updateProperties
        {   progress = count / total_frames,
            text  = string.format("Rendering %d/%d, Image# %d - %02d:%02d:%02d remains",
                                  (count+1),total_frames,frameID,h,m,math.floor(s))
        }
        count = count + 1
        collectgarbage()
    end
    ----------------------------------------------------------------------------

    self.is_running = false
    if self.is_dead then 
         return true
    end

    progress.total.bar:updateProperties  { progress=0, text="Finished" }
    progress.frame.bar:updateProperties  { progress=0, text="" }

    -- --------------------------------------------------------
    --              Set the kernal node back to its
    --                original number of samples
    -- --------------------------------------------------------
    kernelNode:setPinValue(octane.P_MAX_SAMPLES, self.maxSamples, true)
    clearAnimators()
    if saveImage then
        animate.stop.button:updateProperties { enable=false }
        if(animate.runFFmpeg.check.checked==true and not self.is_cancelled) then 
            animate:createVideo()
        end
    end
end
