
scene = {}

-------------------------------------------------------------------------------------------
-- Sort targets alphanumerically. Snippet from:
-- http://notebook.kulchenko.com/algorithms/alphanumeric-natural-sorting-for-humans-in-lua)
-------------------------------------------------------------------------------------------
local function sortRenderTargets(nodes)

    local function padnum(d)
        local dec, n = string.match(d, "(%.?)0*(.+)")
        return #dec > 0 and ("%.12f"):format(d) or ("%s%03d%s"):format(dec, #n, n)
    end
    
    local function compare(n0, n1) 
        local a, b = n0.name, n1.name
        return a:gsub("%.?%d+",padnum)..("%3d"):format(#b)
             < b:gsub("%.?%d+",padnum)..("%3d"):format(#a)
    end
    
    table.sort(nodes, compare)
end 

--------------------------------------------------------------------------------
-- Find all the render targets and fill in its combo list. Targets without
-- kernels are rejected as these won't render.
--------------------------------------------------------------------------------
function scene:initializeRenderTargets()

    local accepted = {}
    local rejected = {}

    for _, renderTargetNode in ipairs(Render.targets) do
        if renderTargetNode:getInputNode(octane.P_KERNEL) ~= nil then
            if renderTargetNode.name ~= "Old Mesh Preview RenderTarget" then
                table.insert(accepted, renderTargetNode) 
            end
        else
            table.insert(rejected, renderTargetNode.name) 
        end
    end

    if #rejected > 0 then
        local msg1 = ""
        local msg2 = ""
        if #accepted == 0 and #rejected == 1 then
            msg1 = "The sole render target is missing a kernel node. A kernel node\n"
            msg1 = msg1 .."must be attached to "..rejected[1].." before it can be included."
            dialog.critical(msg1, true)
            return false
        else
            if #rejected == 1 then
                msg1 = "The Kernel node is missing from the following render target:"
                msg2 = "        "..rejected[1].."\n\nTo include this target, "
                msg2 = msg2.."attach a kernel node to it and restart the script."
                dialog.info(msg1, msg2)
            else
                rlist = ""
                for _,item in pairs(rejected) do rlist = rlist..item.."\n" end
                msg1 = "Kernel nodes are missing from the following render targets:"
                msg2 = rlist.."\nTo include these targets, attach kernel nodes to them and restart the script."
                dialog.info(msg1, msg2)
            end
            if #accepted == 0 then
                dialog.critical("No render targets with kernel nodes are available.", true)
                return false
            end
        end
    end

    if #accepted == 0 then
        dialog.critical("No render targets with kernel nodes are available.", true)
        return false
    end

    Render.targets = accepted

    -- Sort the render targets alphabetically
    sortRenderTargets(Render.targets)

    -- Initialize renderTarget combo list
    local rtName = {}
    for _,item in pairs(Render.targets) do
        table.insert(rtName, item.name)
    end
    local idx = #rtName > 0  and 1 or 0
    renderTarget.combo:updateProperties { items=rtName, selectedIx=idx }

    return true
end

--------------------------------------------------------------------------------
-- Get the render target for the very first time. Check if there is already
-- a selected target. If so, use it and update the combo list accordingly.
--------------------------------------------------------------------------------
local function once_only_renderTarget_lookup()

    -- Check if there is a selected render target
    local target = octane.project.getSelection()[1]

    if target ~= nil then
        if target.type ~= octane.nodeType.NT_RENDERTARGET then
            target = nil
        end
    end

    local idx = -1
    local currentTarget = ""

    if target ~= nil then
        for i, item in ipairs(octane.project.getSceneGraph():getOwnedItems()) do
            if item == target then
                idx = i 
                currentTarget = item
                break
            end
        end
    end

    if idx ~= -1 then
        for i, item in pairs(Render.targets) do
            if item == currentTarget then
                renderTarget.combo.selectedIx = i
            end
        end
    else
        renderTarget.combo.selectedIx = 1
    end

    -- Make sure the above code doesn't run again
    once_only_renderTarget_lookup = function() end
end

--------------------------------------------------------------------------------
function scene:getRenderTarget()

    once_only_renderTarget_lookup()

    if renderTarget.combo.selectedIx > 0 then
        return renderTarget.combo.selectedIx, Render.targets[renderTarget.combo.selectedIx]
    end
    return -1, nil
end

--------------------------------------------------------------------------------
-- Returns copy of the original scene graph, camera node and rendertarget.
-- This prevents us from modifying the original scene.
--------------------------------------------------------------------------------
function scene:getCopy()

    Render.sceneNode  = nil
    Render.targetNode = nil
    Render.camNode    = nil
    Render.envNode    = nil
    Render.sunDirNode = nil

    local idx, currentTarget = scene:getRenderTarget()
    if idx == -1 then
        return false
    end

    local copyScene = octane.nodegraph.createRootGraph("Project Copy")
    local copyRt    = copyScene:copyFromGraph(octane.project.getSceneGraph(), { currentTarget })[1]
    
    if not copyRt or copyRt.type ~= octane.NT_RENDERTARGET then
        dialog.critical("No render target selected", true)
        return false
    end

    local copyCam = copyRt:getInputNode(octane.P_CAMERA)
    if not copyCam or 
        (copyCam.type ~= octane.NT_CAM_PANORAMIC and copyCam.type ~= octane.NT_CAM_THINLENS) then
        dialog.critical("No camera connected to the render target", true)
        return false
    end

    Render.sceneNode  = copyScene
    Render.targetNode = copyRt
    Render.camNode    = copyCam

    return true
end

--------------------------------------------------------------------------------
-- Normally, the camera position reverts to what it was before running the
-- script since everything has been running as a copy of the scene. 
-- There are times when it would be nice to keep the current preview position
-- when exiting the script. If the keepCamera flag is set, then the orginal
-- position will be updated to the current preview frame setting.
--------------------------------------------------------------------------------
local function updateRealCameraPosition(frame)

    local cp = cameraFile.data[frame]
    local idx, currentTarget  = scene:getRenderTarget()
    -- Use getInputNode in case camera is embedded in a linker node which
    -- would cause an error using getConnectedNode.
    local camNode = currentTarget:getInputNode(octane.P_CAMERA)

    local t = Render.camNode:getPinValue(octane.P_TARGET)
    local p = Render.camNode:getPinValue(octane.P_POSITION)
    local u = Render.camNode:getPinValue(octane.P_UP)

    local x = camNode:getProperties()
    if x.type ~= octane.NT_CAM_PANORAMIC and 
        x.type ~= octane.NT_CAM_THINLENS then
        local msg = "Unable to keep camera position. It appears that the camera\n"
        msg = msg .."is part of a group node or none is attached to the render target.\n"
        dialog.critical(msg, false)
    else
        camNode:setPinValue(octane.P_TARGET,   {t[1],t[2],t[3]})
        camNode:setPinValue(octane.P_POSITION, {p[1],p[2],p[3]})
        camNode:setPinValue(octane.P_UP,       {u[1],u[2],u[3]})
    end
end

--------------------------------------------------------------------------------
function scene:checkForKeepingCameraPosition()
    if keepCamera.check.checked == true  then 
        if keepCamera.check.enable == true  then
            updateRealCameraPosition(keyframe.showFrame.slider.value)
        end
    end
end

--------------------------------------------------------------------------------
-- Get the maximum number of samples for the kernel which will be used
-- as the maximum for previewing or animating.
--------------------------------------------------------------------------------
function scene:getMaxSamples()

    local kernelNode = Render.targetNode:getInputNode(octane.P_KERNEL)
    Render.maxSamples = kernelNode:getPinValue(octane.P_MAX_SAMPLES)

    animate.samples.slider.maxValue = Render.maxSamples

    kernelNode:setPinValue(octane.P_MAX_SAMPLES, Render.maxSamples, true)
    return Render.maxSamples
end

function scene:getDefaultOrcFile()
    local projectPath = Render.projectPath
    local sceneFile = octane.project.getCurrentProject()
    local projectPath = octane.file.getParentDirectory(sceneFile)
    local projectName = octane.file.getFileNameWithoutExtension(sceneFile) .. ".orc"
    return octane.file.join(projectPath, projectName)
end

--------------------------------------------------------------------------------
-- Lookup the default ORC file for the current project and if found, load it.
-- This is constructed by concatenating the project path with a file name
-- consisting of the project name + .orc
--------------------------------------------------------------------------------
local function loadDefaultOrcFile()

    Render.projectPath = Render.projectPath:gsub("\\", "/")
    local orcFile = scene:getDefaultOrcFile()

    cameraFile.editor.text = orcFile

    local file = io.open(cameraFile.editor.text, "r")
    if file ~= nil then
        file:close()

        if cameraFile:load() then
            keyframe:setDefaultSliderValues()
            return true
        else
            local msg1 = "The default Orc file:"
            local msg2 = cameraFile.editor.text.."\n\nfailed to load."
            dialog.info(msg1, msg2)
            return false
        end
    else
        cameraFile.editor.text = ""
    end

    return false
end

--------------------------------------------------------------------------------
--//////////////////////////////////////////////////////////////////////////////
--------------------------------------------------------------------------------

function scene:initialize()

    -- Make sure we've got at least one render target with a kernel node.
    -- What's available is loaded in renderTarget.combo
    if not scene:initializeRenderTargets() then
        return false
    end

    if scene:getCopy() == false then
       return false
    end

    gui:setupCallbacks()
    gui:disable()

    if not loadDefaultOrcFile() then
        cameraFile.open.button.enable = true
        project.path.button.enable = true
        return true
    else
        gui:enable()
    end

    keyframe:initializeData()
    return true
end
