-- -----------------------------------------------------------------------------
-- The widget class is used for creating gui components and appending additional
-- elements to existing components. There are default values built in for
-- each type of component so options only need to be passed when a different
-- value is desired. There are three calls which can be made for widgets:
--             1. create()   2. new()   3. append()
--
-- i. create() instantiates a base object for holdng multiple components.
-- 2. new()    instantiates a component
-- 3. append() will append a component to an existing component
--
-- ------------------------------------------------------
-- Examples of self-contained slider:
--
--       startFrame = widget:new('slider') {{"Start"}}
--       startFrame:append('editor') {text="", width=320}
--
--   The 2 calls above have created objects within startFrame 
--   which could be referenced as follows:
--
--      label      = startFrame.label.text
--      max        = startFrame.slider.maxValuewidget:new
--      editorText = startFrame.editor.text
-- ------------------------------------------------------
-- Examples of base widget with multiple components:
--
--      animate           = widget:create()
--      animate.samples   = animate:new('slider') {{"Samples/px"}, log=true}
--      animate.path      = animate:new('button') {"Input file"}
--      animate.path:append('editor') {text = ""}
--      animate:append('group') { children, border=true }
--
-- The lines above allow for the following references:
--      animate.samples.label.text
--      animate.samples.slider.value
--      animate.path.button.text
--      animate.path.editor.text
--      animate.group
--
-- The new() method can be called in two ways. Using animate.samples above 
-- can be written as:
--      animate.samples = animate:new('slider') {{"Samples/px"}, log=true}
--                           or
--      animate:new('slider','samples') {{"Samples/px"}, log=true}
--
--      In both cases the following components are created:
--         animate.samples.label.text
--         animate.samples.slider.value
--
-- ------------------------------------------------------
-- The arguments of new() and append() are always the type of element to
-- create, i.e. "button", "slider", etc. Label objects can optionally be created
-- for sliders, combo lists and check boxes when listed as the first option in
-- their argument list or passing the values with the keyword: label={}.
-- Here's the usage of the various components:
--
-- new('label' ) {" ",...}              Opt: width, height, x, y
-- new('button') {buttonText,...}       Opt: width, height, x, y, tooltip
-- new('slider') {label={" ",...}, ...} Opt: value, min,max,step,log,width,height,x,y
-- new('numeric'){label={" ",...}, ...} Opt: value, min,max,step,log,width,height,x,y
-- new('check')  {label={" ",...},...}  Opt: text,width,height,checked,x,y
-- new('combo')  {label={" ",...},items={,,,,} ...}  Opt: width, height,x,y
-- new('editor') {...}  Opt: text,width,height,x,y,enable
--
-- For screen groups, the needed children objects are built in row and column
-- order. The number of items within each row of a given group must match.
-- In the new() and append() function, children must always be the first argument
-- in the parameter list. The following shows a 3 row, 2 column screen group:
--     children =
--    {
--        { animate.startFrame.label, animate.startFrame.slider },
--        { animate.endFrame.label,   animate.endFrame.slider },
--        { animate.step.label,       animate.step.slider }
--    }
--    animate:append('group'){ children, text="Frame Range", border=true}
--
-- Option for 'group': text,border,padding,inset,center,x,y,debug
-- ---------------------------------------------------------------------------

local methods =
{
    label = function(self, argv)

        if type(argv[1]) == "table" then
            argv = argv[1]
        end
        -------------------------------------
        local text = ""
        if type(argv[1]) == "string" then
            text = argv[1]
        else
            argv = {}
        end
        -------------------------------------
        self.label = octane.gui.create
        {
            type   = octane.gui.componentType.LABEL,
            text   = text,
            width  = argv.width or 100,
            height = argv.height or 24,
            x      = argv.x or 0,
            y      = argv.y or 0
        }
        return self
    end,
    -----------------------------------------------------

    button = function(self, argv)

        local text = ""
        if type(argv[1]) == "string" then
            text = argv[1]
        end
        if argv.tooltip ~= nil then
            if widget.hideToolTips then
                argv.tooltip = nil
            end
        end
        ---------------------------------
        self.button = octane.gui.create
        {
            type    = octane.gui.componentType.BUTTON,
            text    = argv.text or text or  "",
            width   = argv.width or 80,
            height  = argv.height or 30,
            tooltip = argv.tooltip or "",
            x       = argv.x or 0,
            y       = argv.y or 0
        }
        return self
    end,
    -----------------------------------------------------

    slider = function(self, argv)

        self.slider = octane.gui.create
        {
            type        = octane.gui.componentType.SLIDER,
            value       = argv.value or 1,
            minValue    = argv.min or 1,
            maxValue    = argv.max or 3,
            step        = argv.step or 1,
            logarithmic = argv.log or false,
            width       = argv.width or 400,
            height      = argv.height or 24,
            x           = argv.x or 0,
            y           = argv.y or 0
        }
        -------------------------------------
        if argv.label then
            self:append('label'){argv.label}
        elseif type(argv[1]) == "table" then
            self:append('label')(argv)
        end
        return self
    end,
    -----------------------------------------------------

    numeric = function(self, argv)

        self.numeric = octane.gui.create
        {
            type     = octane.gui.componentType.NUMERIC_BOX,
            width    = argv.width or 100,
            height   = argv.height or 24,
            maxValue = argv.max or 256000,
            minValue = argv.min or 0,
            step     = argv.step or 100,
            value    = argv.value or 0,
            x        = argv.x or 0,
            y        = argv.y or 0
        }
        -------------------------------------
        if argv.label then
            self:append('label'){argv.label}
        elseif type(argv[1]) == "table" then
            self:append('label')(argv)
        end
        return self
    end,
    -----------------------------------------------------

    check = function(self, argv)

        self.check = octane.gui.create
        {
            type   = octane.gui.componentType.CHECK_BOX,
            text   = argv.text or "",
            width  = argv.width or 80,
            height = argv.height or 20,
            checked = argv.checked or false,
            x       = argv.x or 0,
            y       = argv.y or 0
        }
        -------------------------------------
        if argv.label then
            self:append('label'){argv.label}
        elseif type(argv[1]) == "table" then
            self:append('label')(argv)
        end
        return self
    end,
    -----------------------------------------------------

    combo = function(self, argv)

        self.combo = octane.gui.create
        {
            type  = octane.gui.componentType.COMBO_BOX,
            name  = argv.name or "",
            items = argv.items or {},
            width = argv.width or 120,
            height = argv.height or 24,
            x      = argv.x or 0,
            y      = argv.y or 0,
            selectedIx = 1
        }
        -------------------------------------
        if argv.label then
            self:append('label'){argv.label}
        elseif type(argv[1]) == "table" then
            self:append('label')(argv)
        end
        return self
    end,
    -----------------------------------------------------

    editor = function(self, argv)

        local text = ""
        if type(argv.text) == "string" then
            text = argv.text
        end
        -------------------------------------
        self.editor = octane.gui.create
        {
            type   = octane.gui.componentType.TEXT_EDITOR,
            text   = text,
            width  = argv.width or 400,
            height = argv.height or 30,
            enable = argv.enable or false,
            x      = argv.x or 0,
            y      = argv.y or 0
        }
        return self
    end,
    -----------------------------------------------------

    tabs = function(self, argv)

        self.tabs = octane.gui.create
        {
            type     = octane.gui.componentType.TABS,
            children = argv.children,
            header   = argv.headers,
            x        = argv.x or 0,
            y        = argv.y or 0
        }
        return self
    end,
    -----------------------------------------------------

    progressBar = function(self, argv)
        self.bar =  octane.gui.create
        {
            type   = octane.gui.componentType.PROGRESS_BAR,
            text   = argv.text or "",
            width  = argv.width or 400,
            height = argv.height or 25,
            x      = argv.x or 0,
            y      = argv.y or 0
        }
        return self
    end,
    -----------------------------------------------------

    group = function(self, argv)

        -- Build list in proper format for children object of screen group
        --
        local function list_children(guiChildren)

            local rows = #guiChildren
            local cols = #guiChildren[1]
            local children = {}
            for _,row in pairs(guiChildren) do
                for c,col in pairs(row) do
                    table.insert(children, col)
                end
            end
            return children, rows, cols
        end
        ------------------------------------------------------
        local group, g_rows, g_cols = list_children(argv[1])

        self.group = octane.gui.create
        {
            type     = octane.gui.componentType.GROUP,
            children = group,                 -- list of child components
            rows     = g_rows,                -- number of rows in group grid
            cols     = g_cols,                -- number of cols in group grid
            text     = argv.text or "",       -- text to display at top of group
            border   = argv.border or false,  -- boolean flag to show border
            padding  = argv.padding or {0,0}, -- internal padding in each cell
            inset    = argv.inset or {0,0},   -- inset of the group component
            centre   = argv.center or argv.centre or false,  -- center the group
            debug    = argv.debug or false,  -- true to show the outlines of the group, handy
            x        = argv.x or 0,
            y        = argv.y or 0
        }
        return self
    end
}

--------------------------------------------------------------------------------

 widget = {

    hideToolTips = false,

    setup = function(self, obj)
               obj = obj or {}
               setmetatable(obj, self)
               self.__index = self
               return obj
           end,
    -----------------------------------------------------

    create  = function(self)
              local obj = self:setup()
              obj.method = methods
              return obj
          end,
    -----------------------------------------------------

    new = function (self, key, name)
              local obj = self:create()
              if name ~= nil then
                  return function(...)
                      self[name] = obj.method[key](obj, ...)
                      return self[name]
                  end
              else
                  return function(...)
                      return obj.method[key](obj, ...)
                  end
              end
          end,
    -----------------------------------------------------

    append = function(self, key)
             return function(...)
               return self.method[key](self, ...)
             end
          end
}
--------------------------------------------------------------------------------


